#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import configparser
import os
from .common import p_error, p_warning

class Configuration:
    def __init__(self, defaultvalues = {}):
        self._defaultvalues = defaultvalues
        self._reader = configparser.ConfigParser()

    def read(self, filename, onlyfirst = False, allownodefault = False):
        filesread = []

        # if filename is not an array, convert it to an array
        if not isinstance(filename, list):
            filename = [ filename ]
        
        # read the configuration files but if onlyfirst is True, only the first one is read
        for _filename in filename:
            if os.path.isfile(_filename):
                self._reader.read(_filename)
                filesread.append(_filename)
                if onlyfirst:
                    break

        # Check the different keys in the default values and convert the obtained values to the same type
        _result = {}
        for section, secconfig in self._defaultvalues.items():
            if section not in self._reader:
                continue
            if section not in _result:
                _result[section] = {}
            readconfig = self._reader[section]
            for key in secconfig:

                # If it is not a tuple, convert it to a tuple
                if not isinstance(secconfig[key], tuple) or isinstance(secconfig[key], list):
                    secconfig[key] = (secconfig[key], lambda x: x)

                # Get the values of the tuple
                the_default = secconfig[key][0]
                the_type = type(the_default)
                the_function = secconfig[key][1]
                if not callable(the_function):
                    the_type = the_function
                    the_function = lambda x: x

                    # If there is an additional parameter, it should be the function
                    if len(secconfig[key]) > 2:
                        the_function = secconfig[key][2]

                # Make a last check to see if the function is callable
                if not callable(the_function):
                    p_warning("The third parameter of the tuple should be a function; ignoring it")
                    the_function = lambda x: x

                # If the key is in the configuration file, extract it, using the same type as the default value
                if key in readconfig:
                    try:
                        if the_type == bool:
                            _result[section][key] = readconfig.getboolean(key, the_default)
                        elif the_type == int:
                            _result[section][key] = readconfig.getint(key, the_default)
                        elif the_type == float:
                            _result[section][key] = readconfig.getfloat(key, the_default)
                        else:
                            _result[section][key] = readconfig.get(key, the_default)
                    except ValueError:
                        p_error("invalid value for {} in {}".format(key, _filename))
                        _result[section][key] = secconfig[key]
                else:
                    _result[section][key] = the_default

                # Now call the function to correct the value
                _result[section][key] = the_function(_result[section][key])

        if allownodefault:
            # Check the values in the configuration file and if there is no a default value, get it from the configuration file
            for section, readconfig in self._reader.items():
                if section not in _result:
                    _result[section] = {}
                for key in readconfig:
                    if key not in _result[section]:
                        _result[section][key] = readconfig[key]

        return (filesread, _result)


#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import sys
from .common import *
from .storage import Storage
import json
from .rawdata import RawData
from .dataseries import DataSeries
from .version import VERSION
from uuid import UUID
from tqdm import tqdm
import argparse
import os
import xlsxwriter
import math

def dataset_to_barchart(data, lower = None, upper = None):

    if data is None:
        return ""

    barchart_graphics = [
        '\u2581',
        '\u2582',
        '\u2583',
        '\u2584',
        '\u2585',
        '\u2586',
        '\u2587',
        '\u2588',
    ]
    if upper is None:
        upper = max(data)
    if lower is None:
        lower = min(data)
    if lower == upper:
        return barchart_graphics[0] * len(data)

    res = ""
    for d in data:
        v = (d - lower) / (upper - lower)
        p = round(v * (len(barchart_graphics) - 1))
        if p < 0:
            p = 0
        if p >= len(barchart_graphics):
            p = len(barchart_graphics) - 1
        res = res + barchart_graphics[p]
    return res

def correctArguments(args, beginTime, endTime):
    # Convert the thresholds to bytes
    args.threshold_disk = toBytes(args.threshold_disk)
    if args.threshold_disk is None:
        p_error("invalid expression for disk threshold")
        sys.exit(1)

    args.threshold_nic = toBytes(args.threshold_nic)
    if args.threshold_nic is None:
        p_error("invalid expression for disk threshold")
        sys.exit(1)

    if args.fromdate is not None:
        args.fromdate = toDate(args.fromdate, beginTime, endTime)
        p_debugv("fromdate: ", args.fromdate)
        if args.fromdate is None:
            p_error("invalid expression for fromdate")
            sys.exit(1)

    if args.todate is not None:
        args.todate = toDate(args.todate, beginTime, endTime)
        p_debugv("todate: ", args.todate)
        if args.todate is None:
            p_error("invalid expression for todate")
            sys.exit(1)

    return args

class OutputFile:
    def __init__(self, fname, overwrite = False) -> None:
        self._closed = False
        if (fname is None) or (fname == "-"):
            self.f = sys.stdout
        else:
            # If file exists abort
            if not overwrite and os.path.isfile(fname):
                raise Exception("file '{}' already exists".format(fname))
            try:
                self.f = open(fname, "w")
            except:
                raise Exception("error opening file '{}'".format(fname))
    def print(self, data):
        if self._closed:
            raise Exception("file already closed")
        self.f.write(data)
    def println(self, data):
        self.print(data + "\n")
    def close(self):
        if self._closed:
            raise Exception("file already closed")
        self._close = True
        self.f.close()

def osidle_analysis():
    parser = argparse.ArgumentParser(allow_abbrev=False)

    parser.add_argument("-q", "--quiet", action="store_true" , help="use the quiet mode (suppress any output except from the resulting analysis, either from the file or stdout)", default=False, dest="quiet")
    parser.add_argument("-r", "--remove-unknown", action="store_true" , help="remove VMs without valid data in the interval", default=False, dest="removeunknown")
    parser.add_argument("-O", "--output", dest="outputfile", default="-", help="output to this file (default: stdout)", type=str)
    parser.add_argument("--overwrite", dest="overwrite", action="store_true", help="overwrite output file if already exists")
    parser.add_argument("-W", "--custom-file", dest="customoptions", default=None, help="file with custom options for specific vm IDs", type=str)
    parser.add_argument("-F", "--from", dest="fromdate", type=str, default="begin", help="make the analysis from this date. Valid expressions are [<reference>-]<count>[suffix], where reference is a choice between the keywords 'now', 'begin', 'end', 'lastweek'... (if ommited, will be set to 'now'), and <suffix> may be a choice between 's', 'm', 'h', 'd', 'w', 'M' and 'y'. Default: 'begin' (i.e. beggining of the data)")
    parser.add_argument("-T", "--to", dest="todate", help="make the analysis up to this date. The format is the same than '--from' parameter. Default: 'now'", type=str, default="now")
    parser.add_argument("-d", "--database", dest="database", help="database to use", default="/var/lib/osidled/osidled.db")
    parser.add_argument("-C", "--no-cpu", dest="analysis_cpu", help="do not consider CPU usage for the analysis", action="store_false", default=True)
    parser.add_argument("-D", "--no-disk", dest="analysis_disk", help="do not consider DISK usage for the analysis", action="store_false", default=True)
    parser.add_argument("-N", "--no-network", dest="analysis_nic", help="do not consider NETWORK usage for the analysis", action="store_false", default=True)
    parser.add_argument("--threshold-disk", dest="threshold_disk", help="threshold for the disk usage (in units/second); accepts suffix B, K, M, G (default is Bytes)", type=str, default="4K")
    parser.add_argument("--threshold-nic", dest="threshold_nic", help="threshold for the nic usage (in units/second); accepts suffix B, K, M, G (default is Bytes)", type=str, default="4K")
    parser.add_argument("-f", "--format", dest="format", help="output format", choices = ["json", "csv", "excel", "shell"], default="json")
    parser.add_argument("-p", "--pretty", dest="pretty", help="pretty print the output", action="store_true", default=False)
    parser.add_argument("-l", "--level", dest="level", help="level of detail severity for the analysis", choices = ["hard", "medium", "soft", "softer"], default="soft")
    parser.add_argument("-S", "--summarize", dest="summarize", help="summarize the usage for the VMs", action="store_true", default=False)
    parser.add_argument("-U", "--under-threshold", dest="threshold_summarize", help="summary will show only VMs under this threshold of puntuation (from 0 to 10)", type=int, default=3)
    parser.add_argument("-o", "--sort", dest="sort", help="Sort the output descending by estimated usage of the VM", action="store_true", default=False)
    parser.add_argument("-i", "--vm-id", dest="vmids", action="append", help="id of the VMs to be analyzed (can appear multiple times)", type=str)
    parser.add_argument("-v", "--verbose", dest="verbose", help="verbose", action="store_true", default=False)
    parser.add_argument("--full-report", dest="fullreport", help="this is a shorthand for --include-eval-data, --include-eval-data-graph and --include-stats", action="store_true", default=False)
    parser.add_argument("-vv", "--verbose-more", dest="verbosemore", help="verbose more", action="store_true", default=False)
    parser.add_argument("--include-eval-data", dest="include_eval_data", action="store_true", default=False, help="include data used for evaluation")
    parser.add_argument("--include-eval-data-graph", dest="include_eval_data_graph", action="store_true", default=False, help="include data used for evaluation but using a graphical representation (uses utf-8 encoding)")
    parser.add_argument("--include-stats", dest="include_stats", help="include the stats in the output", action="store_true", default=False)
    parser.add_argument("--dump-data", dest="dumpdata", action="store_true", default=False, help="dump the data that would be used for the analysis and finalizes")
    parser.add_argument("--info", dest="info", action="store_true", default=False, help="get the information about the data available and exit")
    parser.add_argument("--vmlist", dest="vmlist", action="store_true", default=False, help="if getting information, dump the list of available VMs")
    parser.add_argument('--version', action='version', version=VERSION)

    args = parser.parse_args()

    if not args.quiet:
        if args.verbose:
            setVerbose(1)
        if args.verbosemore:
            setVerbose(2)

    # Connect to the database (if possible)
    storage = Storage(args.database)
    storage.connect()

    # Get the begin and end time of the data
    p_debugv("getting information from the database")
    beginTime = storage.getmint()
    endTime = storage.getmaxt()

    # Correct the arguments using the keywords and special values
    if args.fullreport:
        args.include_eval_data = True
        args.include_eval_data_graph = True
        args.include_stats = True
        
    args = correctArguments(args, beginTime, endTime)

    # Get information
    if args.info:
        print("Information about the data available:")
        print("  - first entry:", beginTime)
        print("  - last entry:", endTime)
        print("  - available vms:", len(storage.getvms()))
        if args.vmlist:
            print("  - available vm ids:", "\n      ".join(["", *storage.getvms()]))
        sys.exit(0)

    # Prepare the output file
    if args.format != "excel":
        try:
            p_debugv("Opening output file {}".format(args.outputfile))
            output_file = OutputFile(args.outputfile, args.overwrite)
        except Exception as e:
            p_error("Error: {}".format(str(e)))
            sys.exit(1)
    else:
        if args.outputfile == "-":
            args.outputfile = "/dev/stdout"

        if args.outputfile == "/dev/stdout":
            p_error("Excel output format is not supported for stdout")
            sys.exit(1)

        if not args.overwrite and os.path.isfile(args.outputfile):
            p_error("file '{}' already exists".format(args.outputfile))
            sys.exit(1)

        try:
            workbook = xlsxwriter.Workbook(args.outputfile)
        except Exception as e:
            p_error("Error: {}".format(str(e)))
            sys.exit(1)

    # Now the default options for each of the specific VM is the options passed in the commandline
    parser.set_defaults(**args.__dict__)

    # Store the default args
    defaultargs = args

    # Custom options for the specific VMs
    customoptions = {}

    if args.customoptions is not None:
        with open(args.customoptions, "r") as f:
            optionsvms = [ l.split("#")[0].strip().split(":") for l in f.readlines() ]
            for options in optionsvms:
                if len(options) == 1 or len(options) == 2:
                    uuid = options[0]
                    try:
                        UUID(uuid)
                        validuuid = True
                    except ValueError:
                        p_warning("invalid uuid {}".format(uuid))

                    if validuuid:
                        p_debugv("custom options for {}: {}".format(uuid, options[1]))
                        if len(options) == 1:
                            # This may be a special case to deactivate the analysis
                            customoptions[uuid] = {}
                        else:
                            customargs, errorargs = parser.parse_known_args(options[1].split())
                            customargs = correctArguments(customargs, beginTime, endTime)

                            if len(errorargs) > 0:
                                p_warning("invalid custom options for {}".format(uuid))

                            customoptions[uuid] = customargs

    # Now get the VM ids to deal with
    if args.vmids is not None:
        vms = args.vmids
    else:
        vms = storage.getvms()

    # Obtain the stats for the different VMs
    rawdata = {}

    if not args.quiet:
        pbar = tqdm(total=len(vms), desc="Processing VMs", unit="VMs")
    for vm in vms:
        if not args.quiet and getVerbose() == 0:
            pbar.update(1)
        p_debug("reading entries for vm {}".format(vm))
        vmdata = storage.getvmdata(vm, args.fromdate, args.todate)
        p_debugv("{} entries found".format(len(vmdata)))

        if len(vmdata) == 0:
            continue

        rawdata[vm] = RawData(vmdata, args)

    # Close the progress bar to avoid weird output
    if not args.quiet:
        pbar.close()

    # If only wanted to get the data, dump it an finalize
    if args.dumpdata:
        for vm, data in rawdata.items():
            output_file.println(data.dumpdata(args.format, args.pretty, transform_fnc=lambda x: (vm, *x)))
        sys.exit(0)

    # Now prepare the results for the analysis
    if args.format == 'json' or args.format == 'shell':
        result = {}
    else:
        result = []

    f_stats = []
    f_evaluation = []
    f_data = []
    h_data = []
    h_stats = []
    fmt_stats = []
    fmt_evaluation = []
    fmt_data = []
    h_evaluation = [ "ID", "Overall" ]

    # This is only for excel output
    fmt_evaluation = [ None, "0.00" ]

    # Depending on what we want to get from the analysis, we prepare the different headers and the data
    if args.analysis_cpu:
        if args.include_stats:
            f_stats = [ *f_stats, "stats.cpu.min", "stats.cpu.max", "stats.cpu.mean", "stats.cpu.median" ]
            h_stats = [ *h_stats, "CPU min", "CPU max", "CPU mean", "CPU median" ]
            fmt_stats = [ *fmt_stats, 10, 10, 10, 10 ]
        f_evaluation = [ *f_evaluation, "cpu.score" ]
        h_evaluation = [ *h_evaluation, "P. CPU" ]
        fmt_evaluation = [ *fmt_evaluation, "0.00" ]
        if args.include_eval_data:
            f_data = [ *f_data, "cpu.data", "cpu.cores" ]
            h_data = [ *h_data, "cpu 0-10", "cpu 10-20", "cpu 20-30", "cpu 30-40", "cpu 40-50", "cpu 50-60", "cpu 60-70", "cpu 70-80", "cpu 80-90", "cpu 90-100", "Suggested cores" ]
            fmt_data = [ *fmt_data, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 0 ]
        if args.include_eval_data_graph:
            f_data = [ *f_data, "cpu.graph" ]
            h_data = [ *h_data, "cpu graph" ]
            fmt_data = [ *fmt_data, None ]

    if args.analysis_disk:
        if args.include_stats:
            f_stats = [ *f_stats, "stats.disk.min", "stats.disk.max", "stats.disk.mean", "stats.disk.median" ]
            h_stats = [ *h_stats, "Disk min", "Disk max", "Disk mean", "Disk median" ]
            fmt_stats = [ *fmt_stats, 0, 0, 0, 0 ]
        f_evaluation = [ *f_evaluation, "disk.score" ]
        h_evaluation = [ *h_evaluation, "P. disk" ]
        fmt_evaluation = [ *fmt_evaluation, "0.00" ]
        if args.include_eval_data:
            f_data = [ *f_data, "disk.data" ]
            h_data = [ *h_data, "disk 0-10", "disk 10-20", "disk 20-30", "disk 30-40", "disk 40-50", "disk 50-60", "disk 60-70", "disk 70-80", "disk 80-90", "disk 90-100" ]
            fmt_data = [ *fmt_data, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10 ]
        if args.include_eval_data_graph:
            f_data = [ *f_data, "disk.graph" ]
            h_data = [ *h_data, "disk graph" ]
            fmt_data = [ *fmt_data, None ]
    if args.analysis_nic:
        if args.include_stats:
            f_stats = [ *f_stats, "stats.nic.min", "stats.nic.max", "stats.nic.mean", "stats.nic.median" ]
            h_stats = [ *h_stats, "NIC min", "NIC max", "NIC mean", "NIC median" ]
            fmt_stats = [ *fmt_stats, 0, 0, 0, 0 ]
        f_evaluation = [ *f_evaluation, "nic.score" ]
        h_evaluation = [ *h_evaluation, "P. nic" ]
        fmt_evaluation = [ *fmt_evaluation, "0.00" ]
        if args.include_eval_data:
            f_data = [ *f_data, "nic.data" ]
            h_data = [ *h_data, "nic 0-10", "nic 10-20", "nic 20-30", "nic 30-40", "nic 40-50", "nic 50-60", "nic 60-70", "nic 70-80", "nic 80-90", "nic 90-100" ]
            fmt_data = [ *fmt_data, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10 ]
        if args.include_eval_data_graph:
            f_data = [ *f_data, "nic.graph" ]
            h_data = [ *h_data, "nic graph" ]
            fmt_data = [ *fmt_data, None ]

    # Now get the analysis
    if not args.quiet:
        pbar = tqdm(total=len(rawdata), desc="Analyzing the data", unit="VMs")

    for vm, data in rawdata.items():
        if not args.quiet and (getVerbose() == 0):
            pbar.update(1)

        p_debug("evaluating data for vm {}".format(vm))
        _stats = DataSeries(data, args)

        stats = _stats.stats
        evaluation = _stats.evaluation
        evaluation["cpu"]["cores"] = math.ceil(_stats.stats["cpu"]["max"]) if _stats.stats["cpu"] is not None else 0

        # Calculate the overall score from the evaluation
        # TODO: check whether to use the 2nd value or not (this is more strict)
        scores = []

        if args.analysis_cpu and evaluation["cpu"]["score"] is not None:
            scores.append(evaluation["cpu"]["score"])

        if args.analysis_nic and evaluation["nic"]["score"] is not None:
            scores.append(evaluation["nic"]["score"])

        if args.analysis_disk and evaluation["disk"]["score"] is not None:
            scores.append(evaluation["disk"]["score"])
        
        # If wanted to remove the unknown values, we do it here
        if (len(scores) == 0) and args.removeunknown:
            p_debugv("removing vm {} because it has no valid data".format(vm))
            continue

        # In the "softer" mode, we overweight the maxium score, so that the VM profile is taken into account
        if args.level == "softer":
            if len(scores) > 0:
                scores = [ *scores, max(scores)]

        overall = round(sum(scores) / len(scores), 2) if len(scores) > 0 else 0

        # If wanted to dump the stats, add them to the evaluation
        if args.include_stats:
            evaluation["stats"] = stats

        # Add the graph data to the information just in case it is needed
        evaluation["cpu"]["graph"] = dataset_to_barchart(evaluation["cpu"]["data"], 0, 1)
        evaluation["disk"]["graph"] = dataset_to_barchart(evaluation["disk"]["data"], 0, 1)
        evaluation["nic"]["graph"] = dataset_to_barchart(evaluation["nic"]["data"], 0, 1)

        # Depending on the format, retrieve the data
        if args.format == "json" or args.format == "shell":
            evaluation["overall"] = overall

            if not args.include_eval_data:
                del evaluation["cpu"]["data"]
                del evaluation["disk"]["data"]
                del evaluation["nic"]["data"]
                del evaluation["cpu"]["data2"]
                del evaluation["disk"]["data2"]
                del evaluation["nic"]["data2"]

            result[vm] = evaluation
        else:
            evaluation["cpu"]["data"] = [ "" if x == 0 else x for x in evaluation["cpu"]["data"] ] if evaluation["cpu"]["data"] is not None else [ ]
            evaluation["disk"]["data"] = [ "" if x == 0 else x for x in evaluation["disk"]["data"] ] if evaluation["disk"]["data"] is not None else [ ]
            evaluation["nic"]["data"] = [ "" if x == 0 else x for x in evaluation["nic"]["data"] ] if evaluation["nic"]["data"] is not None else [ ]

            result.append(
                (vm, overall, *get_fields(evaluation, f_evaluation), *get_fields(evaluation, f_data), *get_fields(evaluation, f_stats))
            )

    # Close the progress bar to avoid weird output
    if not args.quiet:
        pbar.close()

    # Finally, dump the data, depending on the format (moreover we'll sort it and summarize the data, depending on the options)
    if args.format == "json" or args.format == "shell":
        if args.summarize:
            result = { k:v for k,v in result.items() if v["overall"] < args.threshold_summarize }

        if args.sort:
            result = dict(sorted(result.items(), key=lambda x: x[1]["overall"], reverse=True))

    else:
        if args.summarize:
            result = [ x for x in result if x[1] < args.threshold_summarize ]

        if args.sort:
            result = sorted(result, key=lambda x: x[1], reverse=True)

    if args.format == "json":
        if args.pretty:
            output_file.println(json.dumps(result, indent=4))
        else:
            output_file.println(json.dumps(result))
    elif args.format == "shell":
        fields = [ "disk.score", "cpu.score", "nic.score", "overall" ]
        if args.include_eval_data_graph:
            fields = [ *fields, "disk.graph", "cpu.graph", "nic.graph" ]
        if args.include_eval_data:
            fields = [ *fields, "cpu.cores" ]
        if args.include_stats:
            fields = [ *fields, 
                "stats.disk.min", "stats.disk.max", "stats.disk.mean", "stats.disk.median", 
                "stats.cpu.min", "stats.cpu.max", "stats.cpu.mean", "stats.cpu.median", 
                "stats.nic.min", "stats.nic.max", "stats.nic.mean", "stats.nic.median" 
            ]
        c = 0

        for vm, _r in result.items():
            output = [
                "ID_{}={}".format(c, vm)
            ]
            _r = [ x if x is not None else "" for x in get_fields(_r, fields) ]
            for i in range(len(fields)):
                output.append("{}_{}={}".format(fields[i].replace(".", "_"), c, _r[i]))
            output_file.println("\n".join(output))
            c += 1
    else:
        headers = [ *h_evaluation ]
        format = [ *fmt_evaluation ]
        if args.include_eval_data or args.include_eval_data_graph:
            headers = [ *headers, *h_data ]
            format = [ *format, *fmt_data ]
        if args.include_stats:
            headers = [ *headers, *h_stats ]
            format = [ *format, *fmt_stats ]

        if args.format == "excel":
            # Create the excel file
            formats = {}

            worksheet = workbook.add_worksheet()
            for h in headers:
                worksheet.write(0, headers.index(h), h)

            for i in range(len(result)):
                for j in range(len(headers)):

                    c_fmt = None
                    if format[j] is not None:
                        if j not in formats:
                            formats[j] = workbook.add_format({
                                "num_format": format[j]
                            })
                        c_fmt = formats[j]
                        
                    if c_fmt is None:
                        worksheet.write(i+1, j, result[i][j])
                    else:
                        worksheet.write(i+1, j, result[i][j], c_fmt)
            workbook.close()
        else:
            output_file.println(",".join(headers))
            output_file.println("\n".join( [ ",".join([str(x) for x in y]) for y in result ]))        

if __name__ == "__main__":
    osidle_analysis()


#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from .common import *
import math

def _reduce(d, func, init = 0):
    try:
        for k, v in d.items():
            init = func(init, v)
        return init
    except:
        for v in d:
            init = func(init, v)
        return init

# Calculate the basic stats of a serie of values with frequencies
# @param _data: series of rows of tuples where one of the columns is the amount of samples with the same value (set in other column); i.e. (samples, v1, v2, v3, ...)
# @param tpos: column of the amount of samples
# @param vpos: column of the value to calculate the stats
def basicstats(_data, tpos = 0, vpos = 1):
    values = [ (x[tpos], x[vpos]) for x in _data ]
    values.sort(key=lambda x: x[1])

    count = sum([ x[0] for x in values ])
    if count == 0:
        return None

    amount = sum([ x[0] * x[1] for x in values ])
    meanval = amount / count
    mpoint = count // 2
    mpos = 0

    for median_point in values:
        mpos += median_point[0]
        if mpos > mpoint:
            break

    deviation = math.sqrt(sum([ (x[1] - meanval)**2 for x in values ]) / (count - 1))

    if values[-1][1] == values[0][1]:
        pct_deviation = 0
    else:
        pct_deviation = deviation / (values[-1][1] - values[0][1])

    return {
        "min": values[0][1],
        "max": values[-1][1],
        "mean": meanval,
        "deviation": deviation,
        "pct_deviation": pct_deviation,
        "median": median_point[1],
    }

# Function analyzes the percentages of appearance of the values in a dataset, in each of a set of fragments of the possible space of values
#  The idea is to get a dataset that contains values within a certain range of values, then identify a range of possible values and to divide
#  such range in N fragments. Then count the number of times that the values of the dataset fall into each fragment.
#
#  e.g.: have the usage of CPU in each second of a period. The values are between 0 and 100. The range is 0-100% and if we wanted 4 fragments, we'll
#        have 4 fragments of 0-25%, 25-50%, 50-75% and 75-100% of the range. Then we'll analyze each value of the dataset and count how many times
#        it falls into each fragment. Finally, we'll divide per the total amount, to get the percentage of times of appearance of the values in each 
#        fragment.
#
# - The function is made somehow general, so that we could also cluster values in the dataset (e.g. (5 times, value 10), (8 times, value 20), etc.);
#   moreover the number of fragments can be set as parameter.
# - The range is provided as a parameter, and if values get out of this range, they will be absorbed by the borders of the range.
#
# @param data: dataset to analyze (an array of tuples, where there is an element that contains the values and the amount of times that the value appears)
# @param tpos: the column used to the times that a value is repeated (if None, each value will be counted as 1 time)
# @param vpos: the column used to the value to analyze

def _fragments(_data, tpos = None, vpos = 0, low = None, up = None, nfragments = 4):
    if len(_data) == 0:
        return None

    if tpos is None:
        serie = [ (1, x[vpos]) for x in _data ]
    else:
        serie = [ (x[tpos], x[vpos]) for x in _data ]

    if up is None:
        up = max([ x[1] for x in serie ])
    if low is None:
        low = min([ x[1] for x in serie ])

    w = (up - low) / nfragments
    total = 0
    fragments = [0] * nfragments
    for t, v in serie:
        p = int((v - low) // w)

        # Make the borders to absorve the values
        if p < 0: p = 0
        if p >= nfragments: p = nfragments - 1
        total += t
        fragments[p] += t

    result = [ t/total for t in fragments ]
    return result

# DEPRECATED:
# This method converts a series that has a repetition of values into a series that has a different repetition of values, by making the value proportional to the new repetition rate;
#   Column "t" is the repetition amount and the total amount for such repetition is in column "c"; the idea is to make the series continuous and to make the new distribution of values
#   i.e. ((40,100), (30, 90), (50, 200)) can be converted into ((30,75), (30, 25+60), (30, ...) ...)
def _cluster(serie, t, c, tsize, skip=0):
    n_serie = []
    p_e = 0
    p_disk_e = 0

    for row in serie:
        e = row[t]
        disk = row[c]
        disk_e = disk / e

        # Calculate the amount for the first block (if there are remaining values to clusterize)
        if p_e != 0:

            # No esta bien calculado: el nuevo e es tsize o p_e + e como maximo: es p_e + max(e - p_e, tsize - p_e) )

            n_e = min(e - p_e, tsize - p_e)
            n_disk = disk_e * n_e + p_disk_e * p_e
            n_serie.append((n_e + p_e, n_disk))
            e = e - n_e

        # Now the central amount
        n = int(e // tsize)
        n_serie.extend([(tsize, disk_e * tsize)] * n)

        # Calculate the rest
        p_e = e - n * tsize
        p_disk_e = disk_e

    if p_e > 0:
        n_serie.append((p_e, p_disk_e * p_e))

    return n_serie

# The idea of this function is to make a filter to the data, so that if a timestep the value saturates (i.e. gets upper to a threshold), the value is set to the threshold
#   and the remaining amount is added to the next timestep, so that the value is not lost. But the amount added is not the whole amount; instead, it is minored by the
#   an epsilon value.
# The underlying use case is the following:
#   - a VM downloads 10 Gb in just 1 minute, but this is because of the network (either the local endpoint or the remote endpoint)
#   - the same VM may have been downloaded the same amount of data in 1 hour, because the network was slower.
#   - both VMs have been used. If the threshold was 8Kbps, both VMs pass the threshold but the first has a lower score because it was downloading less time.
# Using this filter, the evaluation mechanism tries to be more fair. If the epsilon is too big, the big transferences will be overrated while the small ones will be underrated.
#   * At the end, this is a subjective matter, bucause the actual fact is that the first VM has used the network less time than the second one. So the second VM has been "used"
#     more time than the first one.
def _filter_saturation(serie, t, c, threshold, eps = 0.5):
    for c_s in range(0, len(serie)):
        e = serie[c_s][t]
        disk = serie[c_s][c]

        if disk > threshold:
            diff_disk = disk - threshold
            if (c_s < (len(serie) - 1)):
                e_next = serie[c_s + 1][t]
                n_val = ((serie[c_s + 1][c] * e_next) + (diff_disk * e * eps)) / e_next
                serie[c_s + 1][c] = n_val
            serie[c_s][c] = threshold

def _evaluate_fragment(fragment, minval = 0, maxval = 10):
    # Calculate the points according to the position in the deciles os quartiles
    #   > in quartiles will have 4 pct of usage in 0-25%, 25-50%, 50-75% and 75-100%; 
    if fragment is None:
        return None

    result = 0
    max_p = len(fragment)
    for p in range(0, max_p):
        result += p * fragment[p]
    result = round(maxval * result / (max_p - 1), 2)
    return result

class DataSeries:
    def __init__(self, rawdata, args):
        self._params = {
            "threshold_disk": args.threshold_disk,
            "threshold_nic": args.threshold_nic,
            "level": args.level,
        }

        # Now convert the data into a format ready to be analyzed (frequency, cpu (seconds/second), disk (bytes/second), nic(bytes/second))
        #   TODO(N): consider dividing the x["tcpu"] per ncpus to get the "for analysis" cpu time
        #       ANSWER: NO, because we want to be able to show the data; the "for analysis" cpu time is an internal value
        _data = rawdata.data
        self._data_series = [ [x["e"], x["tcpu"]/x["e"], x["tdisk"]/x["e"], x["tnic"]/x["e"]] for x in _data ]
        # n_data_series = _cluster(self._data_series, 0, 2, 60)
        self._vminfo = {
            "ncpu": rawdata.ncpu,
            "nnic": rawdata.nnic,
            "ndisk": rawdata.ndisk,
        }

        self._stats = None 
        self._evaluation = None

    @property
    def stats(self):
        if self._stats is None:
            self._stats = self._calculate_stats()
        return self._stats

    @property
    def evaluation(self):
        if self._evaluation is None:
            self._evaluation = self._evaluate_stats(self.stats)
        return self._evaluation


    def _evaluate_stats(self, stats):
        eps = 0.75
        if self._params["level"] == "medium":
            eps = 0.75
            ncpu = max(math.ceil(stats["cpu"]["max"]) if stats["cpu"] is not None else 1, 1)
            diskdata = stats["disk"]["median"] if stats["disk"] is not None else None
            nicdata = stats["nic"]["median"] if stats["nic"] is not None else None
            cpudata = stats["cpu"]["median"] if stats["cpu"] is not None else None
        elif self._params["level"] == "hard":
            eps = 0.25
            ncpu = self._vminfo["ncpu"]
            diskdata = stats["disk"]["min"] if stats["disk"] is not None else None
            nicdata = stats["nic"]["min"] if stats["nic"] is not None else None
            cpudata = stats["cpu"]["min"] if stats["cpu"] is not None else None
        elif self._params["level"] == "softer":
            eps = 0.85
            ncpu = 1
            diskdata = stats["disk"]["min"] if stats["disk"] is not None else None
            nicdata = stats["nic"]["min"] if stats["nic"] is not None else None
            cpudata = stats["cpu"]["min"] if stats["cpu"] is not None else None
        else:
            eps = 0.85
            ncpu = 1
            diskdata = stats["disk"]["mean"] if stats["disk"] is not None else None
            nicdata = stats["nic"]["mean"] if stats["nic"] is not None else None
            cpudata = stats["cpu"]["mean"] if stats["cpu"] is not None else None

        # Now we'll filter the disk information and the network to mitigate the impact of saturation
        _filter_saturation(self._data_series, 0, 2, self._params["threshold_disk"], eps)
        _filter_saturation(self._data_series, 0, 3, self._params["threshold_nic"], eps)

        # Calculate the decile distribution for each value
        decile = { 
            "cpu": _fragments(self._data_series, 0, 1, 0, ncpu, 10),
            "disk": _fragments(self._data_series, 0, 2, 0, self._params["threshold_disk"], 10),
            "nic": _fragments(self._data_series, 0, 3, 0, self._params["threshold_nic"], 10),
        }

        evaluation = {
            "disk": {
                "score": _evaluate_fragment(decile["disk"]),
                "data": decile["disk"],
                "score2": round(min(10 * diskdata / self._params["threshold_disk"], 10), 2) if diskdata is not None else None,
                "data2": diskdata
            },
            "nic": {
                "score": _evaluate_fragment(decile["nic"]),
                "data": decile["nic"],
                "score2": round(min(10 * nicdata / self._params["threshold_nic"], 10), 2) if nicdata is not None else None,
                "data2": nicdata
            },
            "cpu": {
                "score": _evaluate_fragment(decile["cpu"]),
                "data": decile["cpu"],
                "score2": round(min(10 * cpudata / ncpu, 10), 2) if cpudata is not None else None,
                "data2": cpudata
            }
        }
        return evaluation
            
    # This function obtains some stats for the data sample
    def _calculate_stats(self):
        # Get the stats
        result = {
            "cpu": basicstats(self._data_series, tpos = 0, vpos = 1),
            "disk": basicstats(self._data_series, tpos = 0, vpos = 2),
            "nic": basicstats(self._data_series, tpos = 0, vpos = 3),
        }
        return result



from .common import *
import argparse
from .version import VERSION
from .storage import Storage, remove_unneeded_data
import shutil
import os
import datetime
from tqdm import tqdm

def osidle_packdb():
    parser = argparse.ArgumentParser(allow_abbrev=False)

    parser.add_argument("-q", "--quiet", action="store_true" , help="use the quiet mode (suppress any output except from the resulting analysis)", default=False, dest="quiet")
    parser.add_argument("-y", "--yes", action="store_true", default=False, dest="force", help="answer yes to all questions")
    parser.add_argument("-n", "--no-backup", dest="backupdb", default=True, action="store_false", help="do not backup the database before packing it")
    parser.add_argument("-b", "--backup-filename", dest="backupfile", default=None, help="name of the file in which to store the backup")
    parser.add_argument("-o", "--overwrite", action="store_true", default=False, dest="overwrite", help="overwrite the backup file if it already exists")
    parser.add_argument("-F", "--keep-from", dest="keepfromdate", type=str, default=None, help="discard any sample before this date. Valid expressions are [<reference>-]<count>[suffix], where reference is a choice between the keywords 'now', 'begin', 'end', 'lastweek'... (if ommited, will be set to 'now'), and <suffix> may be a choice between 's', 'm', 'h', 'd', 'w', 'M' and 'y'. Default: 'begin' (i.e. beggining of the data)")
    parser.add_argument("-T", "--keep-to", dest="keeptodate", help="discard any sample after this date. The format is the same than '--keep-from' parameter. Default: 'now'", type=str, default=None)
    parser.add_argument("-d", "--database", dest="database", help="database to use", default="/var/lib/osidled/osidled.db")
    parser.add_argument("-v", "--verbose", dest="verbose", help="verbose", action="store_true", default=False)
    parser.add_argument("-vv", "--verbose-more", dest="verbosemore", help="verbose more", action="store_true", default=False)
    parser.add_argument("-m", "--minimize", dest="minimize", help="minimize the entries in the database", action="store_true", default=False)
    parser.add_argument('--version', action='version', version=VERSION)

    args = parser.parse_args()

    if not args.quiet:
        if args.verbose:
            setVerbose(1)
        if args.verbosemore:
            setVerbose(2)

    # Connect to the database (if possible)
    storage = Storage(args.database)
    storage.connect()

    # Get the begin and end time of the data
    p_debugv("getting information from the database")
    beginTime = storage.getmint()
    endTime = storage.getmaxt()

    p_debugv("begin time: {}".format(beginTime))
    p_debugv("end time: {}".format(endTime))

    if args.keepfromdate is not None:
        args.keepfromdate = toDate(args.keepfromdate, beginTime, endTime)
        p_debugv("keep-fromdate: ", args.keepfromdate)
        if args.keepfromdate is None:
            p_error("invalid expression for keep-fromdate")
            sys.exit(1)

    if args.keeptodate is not None:
        args.keeptodate = toDate(args.keeptodate, beginTime, endTime)
        p_debugv("keep-todate: ", args.keeptodate)
        if args.keeptodate is None:
            p_error("invalid expression for keep-todate")
            sys.exit(1)

    if not os.path.isabs(args.database):
        args.database = os.path.abspath(args.database)

    if not args.backupdb:
        if not args.force:
            if not user_yes_no_query("are you sure that you do not want to backup the database? (it will be modified and this action cannot be undone)".format(args.database), "n"):
                p_info("aborting the operation")
                sys.exit(0)
        else:
            p_debugv("forcing not backing up the database")

    if args.backupdb:
        if args.backupfile is None:
            args.backupfile = "{}/{}-{}".format(os.path.dirname(args.database), datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"), os.path.basename(args.database))

        backupfile = args.backupfile
        try:
            if os.path.exists(backupfile) and args.force:
                p_debugv("forcing overwriting the backup file")
                args.overwrite = True

            if os.path.exists(backupfile) and not args.overwrite:
                args.overwrite = user_yes_no_query("backup file {} already exists. Overwrite it?".format(backupfile), "n")

            if os.path.exists(backupfile) and not args.overwrite:
                p_error("not overwritting backup file {}".format(backupfile))
                sys.exit(1)

            p_info("backing up database to file {}".format(backupfile))
            shutil.copyfile(args.database, backupfile)
        except Exception as e:
            p_error("could not backup database: {}".format(e))
            sys.exit(1)
    else:
        p_info("not backing up database ")
    
    if not args.force:
        if not user_yes_no_query("are you sure that you want to pack the database? (it will be modified and this action cannot be undone)".format(args.database), "n"):
            p_info("aborting the operation")
            sys.exit(0)

    if (args.keepfromdate is not None) or (args.keeptodate is not None):
        rows = storage.delete(args.keepfromdate, args.keeptodate)
        p_info("{} rows deleted".format(rows))

    def prefilter(rowcount):
        global pbar
        if not args.quiet:
            pbar = tqdm(total=rowcount, desc="Processing entries", unit="entries")

    def postfilter():
        if not args.quiet:
            global pbar
            pbar.close()

    def filterdata(vmid, t, data):
        if not args.quiet:
            global pbar
            pbar.update(1)
        result = remove_unneeded_data(data)
        return result

    if args.minimize:
        storage.filterdata(filterdata, prefilter, postfilter)

    storage.vaccuum()
    p_info("database vaccuumed")
    try:
        pass
    except Exception as e:
        p_error("database packing failed: {}".format(e))
        sys.exit(1)
    # newfname = os.path.dirname(args.database) + "/osidled.db.new"